;------------------------------------------------------
; Sol_Asm assembler core
; Copyright (c) 2004-2005, Bogdan Valentin Ontanu.
; All rights reserved.
;------------------------------------------------------


;------------------------------------
; Main command line processing
;------------------------------------
LOCALS

.data
	main_lp_cmd_line	dd	0

	flag_filename_myself	dd	0
	flag_filename_in	dd	0
	flag_filename_out	dd	0
	flag_filename_lst	dd	0
	flag_filename_res	dd	0

	flag_option_help	dd	0
	flag_option_list	dd	0
	flag_option_labels	dd	0
	flag_option_errors	dd	0
	flag_option_files	dd	0
	flag_option_procs	dd	0
	flag_option_strucs	dd	0

.data?
	sz_filename_myself	db	256 dup(?)
	sz_filename_in		db	256 dup(?)
	sz_filename_out		db	256 dup(?)
	sz_filename_lst		db	256 dup(?)
	sz_filename_res		db	256 dup(?)
.code

;---------------------------------
;Command line tokenizer status
;---------------------------------
P1_HUNT_TOKEN		EQU	1
P1_HUNT_SEPARATOR	EQU	2
;---------------------------------
;Modes
;---------------------------------
P1_MODE_DEFAULT		EQU	1
P1_MODE_OPTION		EQU	2

.data
	p1_parse_status	dd	0
	p1_parse_mode	dd	0
.code

Parse_Cmd_Line PROC
	USES	esi,edi
	ARG	@@lpsz_cmd_line:dword

	LOCAL	@@token_start,@@token_size:dword
	LOCAL	@@filename_crt:dword


	ODS1	<"+++Parse Comand Line:">
	ODS		[@@lpsz_cmd_line]


	mov		esi,[@@lpsz_cmd_line]
	mov		[@@token_size],0
	mov		[@@token_start],esi	

	mov		[p1_parse_status],P1_HUNT_TOKEN
	mov		[p1_parse_mode],P1_MODE_DEFAULT

	;-------------------------
	;status byte scanner
	;-------------------------
@@parse_loop:
	mov		al,byte ptr [esi]

	cmp		al,ASCII_SPACE
	jz		@@Is_Space

	cmp		al,"-"
	jz		@@Is_Option

	cmp		al,"/"
	jz		@@Is_Option

	cmp		al,0
	jz		@@Is_End

	jmp		@@Is_Letter

@@parse_next:
	inc		esi
	inc		[@@token_size]
	jmp		@@parse_loop

@@Is_Letter:
	.IF [p1_parse_status]==P1_HUNT_TOKEN
		mov		[@@token_start],esi
		mov		[@@token_size],0
		mov		[p1_parse_status],P1_HUNT_SEPARATOR
	.ENDIF
	jmp	@@parse_next

@@Is_Space:
	.IF [p1_parse_status]==P1_HUNT_SEPARATOR
		Call	On_Have_Token_Cmd,[@@token_start],[@@token_size]
		mov		[p1_parse_status],P1_HUNT_TOKEN	
	.ENDIF
	jmp	@@parse_next

@@Is_Option:
	.IF [p1_parse_status]==P1_HUNT_TOKEN
		mov		[@@token_start],esi
		mov		[@@token_size],0
		mov		[p1_parse_status],P1_HUNT_SEPARATOR
		mov		[p1_parse_mode],P1_MODE_OPTION
	.ENDIF
	jmp	@@parse_next	

@@Is_End:
	;---------------------------------
	; last null is also separator
	;---------------------------------
	.IF [p1_parse_status]==P1_HUNT_SEPARATOR
		Call	On_Have_Token_Cmd,[@@token_start],[@@token_size]
	.ENDIF

	ODS		<offset sz_filename_in>
	ODS1	<"---Parse Comand Line:">
	xor		eax,eax
	ret
ENDP


.data
	cmd_line_file_crt	dd	0
.code
On_Have_Token_Cmd PROC
	USES	esi,edi
	ARG	@@token_start,@@token_size:dword

	ODS1	<"+++On Have Token Cmd">
	Call	Str_Copy_Nr,[@@token_start],offset sz_tmp1,[@@token_size]
	ODS		<offset sz_tmp1>

	.IF [p1_parse_mode]==P1_MODE_DEFAULT
		ODS1	<"Is filename">
		;----------------------------
		; must be a filename :P
		; so copy it to crt location
		;----------------------------
		mov		edi,offset sz_filename_myself
		mov		eax,[cmd_line_file_crt]		
		mov		ebx,256			;filename max size
		mul		ebx
		add		edi,eax
		
		Call	Str_Copy_Nr,[@@token_start],edi,[@@token_size]
		;---------------------------------
		; also mark we have it
		;---------------------------------
		mov		edi,offset flag_filename_myself
		mov		eax,[cmd_line_file_crt]
		shl		eax,2			;*4==flags=dword size
		add		edi,eax
		mov		eax,1
		mov		[edi],eax
		;--------------------------
		; next filename in sequence
		;--------------------------
		mov		eax,[cmd_line_file_crt]
		inc		eax
		.IF eax>4		;max 5 filenames on line
			mov		eax,4
		.ENDIF
		mov	[cmd_line_file_crt],eax

	.ELSEIF	[p1_parse_mode]==P1_MODE_OPTION
		ODS1	<"Is option">
		
		.IF [@@token_size]==2
			;---------------------------------
			;simple options interpreted here
			;---------------------------------
			mov		esi,[@@token_start]
			inc		esi			;skip "-"
			mov		al,byte ptr [esi]
			.IF al=="h"
				mov		[flag_option_help],1		
			.ELSE
				
				Call	Print_Error0,offset sz_err_022,[@@token_start],[@@token_size]
			.ENDIF
		.ELSE
			;-----------------------------------
			;check long tokens
			;-----------------------------------
			mov		esi,[@@token_start]
			inc		esi
			mov		ecx,[@@token_size]
			dec		ecx
			Call	Check_Cmd_Token,esi,ecx
			.IF eax==-1
				Call	Print_Error0,offset sz_err_022,[@@token_start],[@@token_size]
			.ENDIF
		.ENDIF

		mov		[p1_parse_mode],P1_MODE_DEFAULT		
	.ENDIF

	ret
ENDP


;========================================
; Warning:
;	Any error in name length
;	will most likely GPF on parse
;
;
;========================================

;Cmd_Token_STRU STRUC
;	ctkn_name_len		db	?
;	ctkn_name		db	VARLEN dup(?)
;	ctkn_next_link		dd	?
;	ctkn_method		dd	?
;	ctkn_params		dd	?
;ENDS

.data
	cmd_list:
			db	11,"show_labels"
			dd	$+8
			dd	On_Cmd_Show_Labels

			db	10,"show_procs"
			dd	$+8
			dd	On_Cmd_Show_Procs

			db	11,"show_strucs"
			dd	$+8
			dd	On_Cmd_Show_Strucs

			db	10,"show_files"
			dd	$+8
			dd	On_Cmd_View_Files

			db	09,"no_errors"
			dd	$+8
			dd	On_Cmd_No_Errors

			db	08,"out_text"
			dd	$+8
			dd	On_Cmd_Out_Text
			;-----------------------
			;last token must have 
			;null link
			;-----------------------
			db	4,"help"
			dd	0
			dd	On_Cmd_Help
.code


On_Cmd_Show_Procs PROC
	ODS1	<"+++On Cmd Show Procs">
	mov	[flag_option_procs],1
	ret
ENDP

On_Cmd_Show_Strucs PROC
	ODS1	<"+++On Cmd Show Strucs">
	mov	[flag_option_strucs],1
	ret
ENDP

On_Cmd_Show_Labels PROC
	ODS1	<"+++On Cmd Show Labels">
	mov	[flag_option_labels],1
	ret
ENDP

On_Cmd_View_Files PROC
	ODS1	<"+++On Cmd View files">
	mov	[flag_option_files],1
	ret
ENDP

On_Cmd_No_Errors PROC
	ODS1	<"+++On Cmd No Errors">
	mov	[flag_option_errors],1
	ret
ENDP

On_Cmd_Out_Text PROC
	ODS1	<"+++On Cmd Out Text">
	mov	[flag_debug_file],1
	ret
ENDP

On_Cmd_Help PROC
	ODS1	<"+++On Cmd Help">
	mov	[flag_option_help],1
	ret
ENDP

;=============================
;
; Check comand line token
;
;=============================
;Check_Cmd_Token PROTO :dword,:dword
Check_Cmd_Token PROC
	USES	esi,edi
	ARG	@@token_start,@@token_size:dword

	ODS1	<"+++Check Cmd Token:">

	Call	Str_Copy_Nr,[@@token_start],offset sz_tmp1,[@@token_size]
	ODS	<offset sz_tmp1>
	ODV	[@@token_size]

	mov	edi,offset cmd_list

@@loop_tokens:
	mov	esi,[@@token_start]
	mov	ecx,[@@token_size]	

	movzx	edx,byte ptr [edi]	;edx==this token size

	inc	edi			;position on token string start

	;-------------------------
	;fast test length
	;-------------------------
	cmp	edx,ecx			;compare size
	jnz	@@next_token
	

	Call	Str_Cmp_Nz_Token,esi,ecx,edi,edx

	cmp	eax,1
	jz	@@ok_match

@@next_token:
	;-------------------------------------
	; NO match this time
	; --> try next token
	;-------------------------------------
	add	edi,edx		;skip token lenght
	mov	edi,[edi]	;get pointer to next token
	test	edi,edi
	jz	@@no_match

	jmp	@@loop_tokens



@@ok_match:
	ODS1	<"CmdLine Token Match OK">

	add	edi,edx		;add string length
	add	edi,4		;skip next pointer
	mov	eax,[edi]	;get procedure to execute==method
	add	edi,4		;position on param
	;-------------------------
	;Execute Method
	;-------------------------
	pushad
		Call	eax
	popad

	
	;-------------------------
	;return eax=0 for match
	;-------------------------
	xor	eax,eax
	ret	


@@no_match:
	ODS1	<"NO Match Cmd">
	;-------------------------------
	; return eax=-1 for NO match
	;-------------------------------
	mov	eax,-1
	ret
ENDP

