;------------------------------------------------------
; Sol_Asm assembler core
; Copyright (c) 2004-2005, Bogdan Valentin Ontanu.
; All rights reserved.
;------------------------------------------------------

;=================================
; Labels
;=================================

LABEL_STRU STRUC
	label_name	db	128 dup(?)
	label_name_len	dd	?
	label_next	dd	?

	label_type	dd	?
	label_size	dd	?

	label_segment	dd	?
	label_offset	dd	?
ENDS

;---------------------------
;Label types
;---------------------------
LABEL_DATA		EQU	0
LABEL_CODE		EQU	1

LABEL_MAX		EQU	128*1024
LABEl_HASH_MAX		EQU	65538
LABEL_HASH_MASK		EQU	0FFFFh

.data
	label_count	dd		-1
.data?
	labels_base0	LABEL_STRU	LABEL_MAX dup (<?>)

	labels_hash0	dd	65538	dup(?)

.code

Clear_Hash_Labels PROC
	USES	ecx,edi

	ODS1	<"+++Clear Hash Labels">

	mov		edi,offset labels_hash0
	mov		ecx,65537
	mov		eax,-1

@@loop_clear:
	mov		[edi],eax
	add		edi,4
	dec		ecx
	jnz		@@loop_clear

	ret
ENDP

Create_Label PROC
	USES	ecx,ebx,esi,edi
	ARG 	@@token_start,@@token_size:dword
	ARG	@@symbol_type,@@symbol_value:dword
	LOCAL	@@tmp_id:dword
	
	ODS1	<"+++Create Label">

	;Call	Str_Copy_Nr,[@@token_start],offset sz_tmp1,[@@token_size]
	;ODS	<offset sz_tmp1>

	mov	eax,[label_count]
	inc	eax
	mov	[label_count],eax

	cmp	eax,LABEL_MAX
	jc	@@label_ok

	dec	eax
	mov	[label_count],eax
	;-------------------------------------
	; error can not create more labels
	;-------------------------------------
	Call	Print_Error,offset sz_err_006,[@@token_start],[@@token_size]
	mov	eax,-1
	ret

@@label_ok:
	mov	esi,offset labels_base0
	mov	ebx,SIZE LABEL_STRU
	mul	ebx
	add	esi,eax

	push	esi
		add		esi,label_name
		Call	Str_Copy_Nr,[@@token_start],esi,[@@token_size]
	pop	esi
	
	mov	eax,[@@token_size]
	mov	[esi.label_name_len],eax

	mov	eax,[@@symbol_type]
	mov	[esi.label_type],eax

	.IF eax==LABEL_CODE
		mov	[esi.label_size],4
	.ELSE
		.IF [parse_mode]==P2_MODE_DEFINE_B
			mov	[esi.label_size],1
		.ELSEIF [parse_mode]==P2_MODE_DEFINE_W
			mov	[esi.label_size],2
		.ELSEIF [parse_mode]==P2_MODE_DEFINE_D
			mov	[esi.label_size],4
		.ELSE	
			mov	[esi.label_size],4
		.ENDIF
	.ENDIF

	mov	eax,[@@symbol_value]
	mov	[esi.label_offset],eax

	;--------------------------------
	;make hash value
	;--------------------------------
	push	esi

	mov	edi,[@@token_start]
	mov	esi,[@@token_size]
	Call	Make_Hash
	pop	esi
	
	ODVh	ebx
	;-----------------------
	; ebx=hash
	;-----------------------
	shl	ebx,2
	mov	edi,offset labels_hash0
	add	edi,ebx
	mov	eax,[edi]

	.IF eax==-1
		;-----------------------
		;empty hash location
		;-----------------------
		ODS1	<"Add on empty hash">
		mov	ebx,[label_count]
		mov	[edi],ebx
		mov	[esi.label_next],eax
	.ELSE
		;------------------------
		;colision, eax=old guy
		;------------------------
		ODS1	<"Add collission hash">
		mov	ebx,[label_count]
		mov	[edi],ebx
		mov	[esi.label_next],eax
	.ENDIF

	mov	eax,esi
	ret
ENDP




;====================================
; Verify if this Label exists
;====================================
align 4
Check_Label PROC
	USES	esi,edi
	ARG	@@token_start,@@token_size:dword


	ODS1	<"+++Check Label Hash">

	Call	Str_Copy_Nr,[@@token_start],offset sz_tmp1,[@@token_size]
	ODS	<offset sz_tmp1>
	ODV	[@@token_size]

;	mov	eax,-1
;	ret

	mov	ecx,[label_count]
	cmp	ecx,-1
	jz	@@no_labels

	;--------------------------------
	;make hash value
	;--------------------------------
	push	esi
	mov	edi,[@@token_start]
	mov	esi,[@@token_size]
	Call	Make_Hash
	pop	esi
	;-----------------------
	; ebx=hash
	;-----------------------
	ODVh	ebx

	shl	ebx,2
	mov	edi,offset labels_hash0
	add	edi,ebx
	mov	eax,[edi]

	ODS1	<"Label:ID at hash:">
	ODV	eax

	.IF eax==-1
		;--------------------------------
		; lucky: empty hash location
		;--------------------------------
		ODS1	<"Label Not found, empty hash">
		mov	eax,-1
		ret
	.ELSE
		;-----------------------------------------
		; hash hit, we still have to check
		;-----------------------------------------
		ODS1	<"Hash Hit">
		
	@@follow_link:
		ODS1	<"ID tested:">
		ODV	eax	
		
		mov	ebx, SIZE LABEL_STRU
		mul	ebx
		mov	edi,offset labels_base0
		add	edi,eax
		
		mov	eax,[edi.label_name_len]
		cmp	eax,[@@token_size]
		jnz	@@next_link
		
		ODS1	<"Label Length is Equal">
		
		push	edi
		
		mov	esi,[@@token_start]
		add	edi,label_name		
		mov	ecx,[@@token_size]
		
		@@loop_cmp:
			mov	al,byte ptr [esi]
			mov	bl,byte ptr [edi]
			cmp	al,bl
			jnz	@@no_match
			
			inc	edi
			inc	esi
			dec	ecx
			jnz	@@loop_cmp
			
			ODS1	<"Label Found">
			
			pop	edi
			mov	eax,[edi.label_size]	;size 4:code and dd, 2:word,1:byte
			mov	ebx,[edi.label_offset]	;OFFSET
			ret			
			
		@@no_match:
			ODS1	<"Char not match">
			
			pop	edi
			
		@@next_link:
			mov	eax,[edi.label_next]
			
			cmp	eax,-1
			jnz	@@follow_link	
			
			mov	eax,-1
			ret
	.ENDIF

@@no_labels:
	mov	eax,-1
	ret
ENDP



;===============================
; FNV-1 Hash function
;-------------------------------
;In:
;	edi=token_start
;	esi=token_size
;Out:
;	ebx
;==========================
align 4
Make_Hash_Old PROC
	mov	eax,0
	mov	ebx,16777619
@@loop_hash:
	movzx	ecx,byte ptr [edi]

	mul	ebx
	xor	eax,ecx


	inc	edi
	dec	esi
	jnz	@@loop_hash

	mov	ebx,eax
	shr	eax,16
	and	ebx,0ffffh
	xor	eax,ebx
	
	ret
ENDP

Make_Hash PROC
	mov	eax,2166136261		;FNV_Offset_32
	mov	ebx,16777619		;FNV_Prime_32
@@loop_hash:
	movzx	ecx,byte ptr [edi]
	xor	eax,ecx

	mul	ebx

	inc	edi
	dec	esi
	jnz	@@loop_hash

	mov	ebx,eax
	shr	eax,16
	and	ebx,0ffffh
	xor	ebx,eax
	
	;-----------------------
	; return ebx
	;-----------------------
	ret
ENDP

Make_Hash2 PROC
	mov	eax,2166136261		;FNV_Offset_32
	mov	ebx,16777619		;FNV_Prime_32
@@loop_hash:
	movzx	ecx,byte ptr [edi]
	xor	eax,ecx

	;mul	ebx
	
	;---------------------------------
	; a series of shifts and adds 
	; to replace the MUL
	;---------------------------------
	mov	ebx,eax
	shl	ebx,1		;hash shl 1
	add	eax,ebx
	shl	ebx,3		;hash shl 4
	add	eax,ebx
	shl	ebx,3		;hash shl 7
	add	eax,ebx
	shl	ebx,17		;hash shl 24
	add	eax,ebx
	
	inc	edi
	dec	esi
	jnz	@@loop_hash

	;---------------------------
	; fold HASH to 16bits
	;---------------------------
	mov	ebx,eax
	shr	eax,16
	and	ebx,0ffffh
	xor	ebx,eax
	
	;-----------------------
	; return ebx
	;-----------------------
	ret
ENDP
 
;====================================
; Verify if this Label exists
; the very old non-hash version
;====================================
Check_Label1 PROC
	USES	ecx,esi,edi
	ARG		@@token_start,@@token_size:dword

	ODS1	<"+++Check Label">

	;Call	Str_Copy_Nr,[@@token_start],offset sz_tmp1,[@@token_size]
	;ODS	<offset sz_tmp1>
	;ODV	[@@token_size]

	mov	edi,offset labels_base0

	mov	ecx,[label_count]
	cmp	ecx,-1
	jz	@@no_labels

	inc	ecx

@@loop_tokens:
	mov	eax,[edi.label_name_len]
	cmp	eax,[@@token_size]
	jnz	@@next_label

	ODS1	<"Label Length is Equal">

	mov	esi,[@@token_start]
	mov	edx,[@@token_size]

	push	edi

	add	edi,label_name
@@loop_cmp:
	mov	al,byte ptr [esi]		
	mov	bl,byte ptr [edi]
	cmp	al,bl
	jnz	@@no_match
	
	inc	edi
	inc	esi
	dec	edx
	jnz	@@loop_cmp

	pop	edi

	ODS1	<"Label Found">
	mov	eax,[edi.label_size]	;size 4:code and dd, 2:word,1:byte
	mov	ebx,[edi.label_offset]	;OFFSET

	ret

@@no_match:
	;ODS1	<"No match char">	
	pop	edi

@@next_label:	
	;---------------
	; next label
	;---------------
	add	edi,SIZE LABEL_STRU
	dec	ecx
	jnz	@@loop_tokens

@@no_labels:
	ODS1	<"Label Not found">
	mov	eax,-1
	ret
ENDP


;=================================
; Util function
;=================================
.data
	sz_info_label	db	13,10,"Printing Labels:",13,10
			db	"----------------------------------",0

	sz_info_label2	db	13,10
			db	"----------------------------------",13,10,0

	sz_fmt_label	db	13,10,"[%i]%s",09,"Addr:%i",09,"Next:%i",0
.code
Print_Labels PROC
	USES	esi,edi
	LOCAL	@@tmp_cnt:dword

	ODS1	<"+++Print Lables">

	.IF [label_count]==-1
		ret
	.ENDIF

	Call	Print_Text,offset sz_info_label

	mov	[@@tmp_cnt],0
	mov	esi,offset labels_base0
	mov	ecx,[label_count]
	cmp	ecx,-1
	jz	@@end_labels

	inc	ecx

@@loop_labels:
	push	ecx
	push	esi
	
	mov	eax,[esi.label_name_len]
	Call	Str_Copy_Nr,esi,offset sz_tmp1,eax

	mov	eax,[esi.label_offset]
	mov	ebx,[esi.label_next]
	Call	wsprintfA,offset sz_tmp2,offset sz_fmt_label,[@@tmp_cnt],offset sz_tmp1,eax,ebx
	Call	Print_Text,offset sz_tmp2

	pop	esi
	pop	ecx
	;---------------
	; next label
	;---------------
	inc	[@@tmp_cnt]
	add	esi,SIZE LABEL_STRU
	dec	ecx
	jnz	@@loop_labels

@@end_labels:
	Call	Print_Text,offset sz_info_label2

	ret
ENDP
