;------------------------------------------------------
; Sol_Asm assembler core
; Copyright (c) 2004-2005, Bogdan Valentin Ontanu.
; All rights reserved.
;------------------------------------------------------


;====================================
; The MAIN Parser
;====================================


.data
	flag_unresolved		dd	0
	pass_nr			dd	0

	sz_fmt_pass		db	13,10,13,10,"Assemble Pass:%i",0

	bytes_read		dd	0
	bytes_written		dd	0

	lp_out_buff		dd	0

	;-------------------------
	;Output counter and
	;Location counter
	;-------------------------
	out_crt			dd	0
	addr_crt		dd	0

	flag_debug_file		dd	0
	file_handle_debug	dd	0
	file_handle_out		dd	0
	
	sz_filename_debug	db	"out.txt",0
	
	align 4
.code

Parser_Main PROC
	USES	esi,edi
	;---------------------------------------------
	; buffer for output code
	; kind of fixed now :P
	; yes i know it is lame, it will change
	;---------------------------------------------
	Call	GlobalAlloc,GPTR,16*1024*1024	;16-Megabyte max bin output :P
	mov	[lp_out_buff],eax

	.IF [flag_debug_file]==1
		;-----------------------
		;create debug output file
		;-----------------------
		Call	CreateFileA,offset sz_filename_debug,GENERIC_WRITE,0,0,CREATE_ALWAYS,80h,0
		.IF	eax==-1
			ODS1 "ERROR creating output file"
			ret
		.ENDIF
		mov	[file_handle_debug],eax		
	.ENDIF

	;-----------------------
	; Start the Fun
	;-----------------------
	mov	[pass_nr],1

@@loop_pass:

	mov	[out_crt],0
	mov	[addr_crt],0

	;------------------------
	; show pass number
	;------------------------
	Call	wsprintfA,offset sz_tmp1, offset sz_fmt_pass,[pass_nr]
	Call	Print_Text, offset sz_tmp1

	call	Parse_File,offset sz_filename_in
	.IF	eax!=0
		ret
	.ENDIF

	.IF [flag_option_labels]==1
		Call	Print_Labels
	.ENDIF

	.IF [flag_option_procs]==1
		Call	Print_Procs
	.ENDIF

	.IF [flag_option_strucs]==1
		Call	Print_Structures
	.ENDIF

	mov	eax,[pass_nr]
	inc	eax
	mov	[pass_nr],eax

	cmp	eax,2
	jbe	@@loop_pass

	;-----------------------
	;create output file
	;-----------------------
	Call	CreateFileA,offset sz_filename_out,GENERIC_WRITE,0,0,CREATE_ALWAYS,80h,0
	.IF	eax==-1
		ODS1 "ERROR creating output file"
		ret
	.ENDIF
	mov	[file_handle_out],eax

	;---------------------
	;write it to disk
	;---------------------
	call    WriteFile,[file_handle_out],[lp_out_buff],[out_crt],offset bytes_written,0
	call	CloseHandle,[file_handle_out]

	.IF [flag_debug_file]==1
		Call	CloseHandle,[file_handle_debug]
	.ENDIF

	ret
ENDP


;---------------------------------
;Tokenizer level 1 status
;---------------------------------
P2_IGNORE_NEXT			EQU	1
P2_HUNT_SEPARATOR		EQU	2
P2_HUNT_SEPARATOR2		EQU	3	;double quoted
P2_HUNT_SEPARATOR3		EQU	4	;single quoted?
P2_HUNT_TOKEN			EQU	5
P2_HUNT_END_BR			EQU	6	; end of []
P2_HUNT_END_COMMENT2	EQU	7	; /* */
;--------------------------------
;Parser level 2 status
;--------------------------------
P2_MODE_DEFAULT			EQU	0
P2_MODE_INCLUDE			EQU	1
P2_MODE_DEFINE_B		EQU	2
P2_MODE_DEFINE_W		EQU	3
P2_MODE_DEFINE_D		EQU	4
P2_MODE_JMP				EQU	5
P2_MODE_ORG				EQU	6
P2_MODE_DISP			EQU	7
P2_MODE_RETN			EQU	8
P2_MODE_CALL			EQU	9
P2_MODE_POP				EQU	10
P2_MODE_PUSH			EQU	11
P2_MODE_USES			EQU	12
P2_MODE_ARG				EQU	13
P2_MODE_LOCAL			EQU	14
P2_MODE_MOV1			EQU	15
P2_MODE_MOV2			EQU	16
P2_MODE_STRU			EQU	17
;-----------------------------------
;Parser level 3 status
;-----------------------------------
P2_HLL_DEFAULT			EQU	0
P2_HLL_PROC_START		EQU	1
P2_HLL_PROC_BODY		EQU	2
P2_HLL_STRUC_START		EQU	3
P2_HLL_STRUC_BODY		EQU	4
P2_HLL_MACRO_START		EQU	5
P2_HLL_MACRO_BODY		EQU	6
P2_HLL_STRU_START		EQU	7
;-----------------------------------
;Ascii defines
;-----------------------------------
ASCII_CR				EQU	0dh
ASCII_LF				EQU	0ah
ASCII_TAB				EQU	09h
ASCII_SPACE				EQU	20h

;-----------------------------------
;include file level structure
;-----------------------------------
INCLUDE_CTX STRUC
	include_file_name	db	256 dup(?)	;max include filename
	include_line_crt	dd	?
ENDS

.data
	align 4
	char_old			db	0
	char_crt			db	0	

	align 4	
	text_line_crt		dd	0

	parse_hll_mode_old	dd	0
	parse_mode_old		dd	0

	parse_hll_mode		dd	0
	parse_mode			dd	0
	parse_status		dd	0

	include_level		dd	-1

.data?
	;---------------------------------------------
	; context stack for each level of include
	;---------------------------------------------
	include_ctx0	INCLUDE_CTX	2048 dup(<?>)	
.code

;=======================================
; return pointer to THIS 
; include's context
;=======================================
Get_Include_Context PROC
	USES	ebx,edx

	xor		edx,edx
	mov		eax,[include_level]
	mov		ebx,SIZE INCLUDE_CTX
	mul		ebx
	add		eax,offset include_ctx0

	ret
ENDP

.data
	sz_separ1	db	">",0
.code
Parse_File PROC
	USES	esi,edi
	ARG		@@lpsz_file:DWORD

	LOCAL	@@file_handle_crt,@@lp_file_buff,@@file_size:DWORD
	LOCAL	@@token_size,@@token_start,@@parse_count:DWORD

	inc		[include_level]

	ODS1	<"+++Parse file:">
	ODS	[@@lpsz_file]
	ODV	[include_level]
	ODS1	<"---------------">

	;-----------------------------------
	; verbose show the includes parsed
	;-----------------------------------
	.IF [flag_option_files]==1
		Call	Print_Text,offset sz_crlf

		mov	eax,[include_level]
		inc	eax

		@@loop_level:
			push	eax
			Call	Print_Text,offset sz_separ1
			pop	eax
			dec	eax
			jnz	@@loop_level

		Call	Print_Text,[@@lpsz_file]
	.ENDIF


	;----------------------------------
	;setup include context
	;1)copy include name for safe keeping
	;2)make line_crt=0
	;----------------------------------
	Call	Get_Include_Context
	mov	esi,eax
	add	esi,offset include_file_name

	Call	Str_Copy,[@@lpsz_file],esi
	ODS1	<"Include context:">
	ODS	esi

	xor	eax,eax
	inc	eax
	mov	[esi.include_line_crt],eax
	

	;----------------------------------
	; open the input file
	;----------------------------------
	Call	CreateFileA,[@@lpsz_file],GENERIC_READ,0,0,OPEN_EXISTING,80h,0

	.IF	eax==-1
		ODS1	<"ERROR opening input file">
		Call	Print_Text,offset sz_err_002
		Call	Print_Text,[@@lpsz_file]
		dec	[include_level]
		ret
	.ELSE
		ODS1	"OK opening input file"
		ODS	[@@lpsz_file]
	.ENDIF


	mov	[@@file_handle_crt],eax

	Call	GetFileSize,[@@file_handle_crt],0
	mov	[@@file_size],eax

	ODS1	<"File size is:">
	ODV	eax

	add	eax,32
	call	GlobalAlloc,GPTR,eax
	mov	[@@lp_file_buff],eax

	;------------------------------------
	; full read and then close the file
	;------------------------------------
	call    ReadFile,[@@file_handle_crt],[@@lp_file_buff],[@@file_size],offset bytes_read,0
	call	CloseHandle,[@@file_handle_crt]


	;-------------------------------
	; start the real parsing
	;-------------------------------
	mov	esi,[@@lp_file_buff]

	mov	[parse_status],P2_HUNT_TOKEN
	mov	[parse_mode],P2_MODE_DEFAULT	
	mov	[parse_hll_mode],P2_HLL_DEFAULT

	mov	[@@parse_count],0
	mov	[@@token_size],0
	mov	[@@token_start],esi


	mov	[proc_crt],-1
	mov	[stru_crt],-1
	mov	[smem_crt],-1

	xor	eax,eax

	;-----------------------------
	; The byte scan parser ;)
	;-----------------------------
@@parse_loop:
	;-------------------
	;read one char
	;-------------------
	mov	al,byte ptr [esi]
	;----------------------
	; keep a short history
	;----------------------
	mov	bl,[char_crt]
	mov	[char_old],bl
	mov	[char_crt],al

	;--------------------
	; The big CASE
	;--------------------
	cmp	al,ASCII_CR	;CR
	jz	@@Is_CR

	cmp	al,ASCII_LF	;LF
	jz	@@Is_LF

	cmp	al,ASCII_SPACE
	jz	@@Is_Space

	cmp	al,ASCII_TAB
	jz	@@Is_TAB

	cmp	al,";"
	jz	@@Is_Semi

	cmp	al,"/"
	jz	@@Is_Slash

	cmp	al,"*"
	jz	@@Is_Star

	cmp	al,","
	jz	@@Is_Comma

	cmp	al,'['
	jz	@@Is_Left_Br

	cmp	al,']'
	jz	@@Is_Right_Br

	cmp	al,'"'
	jz	@@Is_Quote2

	cmp	al,":"
	jz	@@Is_Letter

	cmp	al,"?"
	jz	@@Is_Letter

	cmp	al,"."
	jz	@@Is_Point

	cmp	al,"="
	jz	@@Is_Equal

	;-------------------------------
	; digits
	;-------------------------------	
	cmp	al,"0"
	jz	@@Is_Digit
	cmp	al,"1"
	jz	@@Is_Digit
	cmp	al,"2"
	jz	@@Is_Digit
	cmp	al,"3"
	jz	@@Is_Digit
	cmp	al,"4"
	jz	@@Is_Digit
	cmp	al,"5"
	jz	@@Is_Digit
	cmp	al,"6"
	jz	@@Is_Digit
	cmp	al,"7"
	jz	@@Is_Digit
	cmp	al,"8"
	jz	@@Is_Digit
	cmp	al,"9"
	jz	@@Is_Digit

	cmp	al,"A"
	jc	@@no_letter
	cmp	al,127
	jnc	@@no_letter

	jmp	@@Is_Letter

@@no_letter:
	jmp	@@Is_Other


@@parse_next:
	inc	esi
	inc	[@@token_size]

	mov	eax,[@@parse_count]
	inc	eax
	mov	[@@parse_count],eax

	cmp	eax,[@@file_size]
	jc	@@parse_loop

	mov	eax,[include_level]
	dec	eax
	mov	[include_level],eax
	
	.IF eax==-1
		Call	Print_Text,offset sz_finish_ok
	.ENDIF

	ODS1	<"---End Parse file---">
	ODS2	<"level:">
	ODV	[include_level]
	ODS1	<"-------------------">

	;------------------------------------------
	; free the memory allocated for the include
	;------------------------------------------
	Call	GlobalFree,[@@lp_file_buff]
	;---------------------
	; return ok
	;---------------------
	xor	eax,eax	
	ret


@@Is_Letter:
	.IF [parse_status]==P2_HUNT_TOKEN
		ODS1	<"Start Token: letter,">
		mov	[@@token_start],esi
		mov	[@@token_size],0
		mov	[parse_status],P2_HUNT_SEPARATOR
	.ENDIF
	jmp	@@parse_next

@@Is_Point:
	.IF [parse_status]==P2_HUNT_TOKEN
		ODS1	<"Start Token: point,">
		mov	[@@token_start],esi
		mov	[@@token_size],0
		mov	[parse_status],P2_HUNT_SEPARATOR		
	.ENDIF
	jmp	@@parse_next

@@Is_Left_Br:
	.IF [parse_status]==P2_HUNT_TOKEN
		ODS1	<"Start Token: Left Bracket,">	
		mov	[@@token_start],esi
		mov	[@@token_size],0
		mov	[parse_status],P2_HUNT_END_BR
		
	.ELSEIF [parse_status]==P2_HUNT_END_BR
		Call	Print_Error,offset sz_err_015,[@@token_start],[@@token_size]	
	.ENDIF

	jmp	@@parse_next


@@Is_Right_Br:
	.IF [parse_status]==P2_HUNT_END_BR
		ODS2 	<"End Token: Right Bracket.">
		mov	eax,[@@token_size]
		inc	eax
		Call	On_Have_Token_Br,[@@token_start],eax
		mov	[parse_status],P2_HUNT_TOKEN
		
	.ELSEIF [parse_status]==P2_HUNT_TOKEN	
		Call	Print_Error,offset sz_err_019,[@@token_start],[@@token_size]
		
	.ENDIF

	jmp	@@parse_next

@@Is_Quote2:
	.IF [parse_status]==P2_HUNT_TOKEN
		ODS1	<"Start Token: Quote2,">	
		mov	[@@token_start],esi
		mov	[@@token_size],0
		mov	[parse_status],P2_HUNT_SEPARATOR2

	.ELSEIF [parse_status]==P2_HUNT_SEPARATOR2
		ODS2 	<"End Token: Quote2.">
		mov	eax,[@@token_size]
		inc	eax
		Call	On_Have_Token_String,[@@token_start],eax
		mov	[parse_status],P2_HUNT_TOKEN		
		
	.ENDIF
	jmp	@@parse_next

@@Is_Digit:
	.IF [parse_status]==P2_HUNT_TOKEN
		ODS1	<"Start Token: Value,">
		mov	[@@token_start],esi
		mov	[@@token_size],0
		mov	[parse_status],P2_HUNT_SEPARATOR
	.ENDIF	

	jmp	@@parse_next


	
@@Is_Space:
	.IF [parse_status]==P2_HUNT_SEPARATOR
		ODS2 	<"End Token: space.">
		Call	On_Have_Token,[@@token_start],[@@token_size]
		mov	[parse_status],P2_HUNT_TOKEN	
	.ENDIF

	jmp	@@parse_next

@@Is_CR:
	;----------------------------------------------
	; this is also a separator if we hunt for it
	;----------------------------------------------
	.IF [parse_status]==P2_HUNT_SEPARATOR
		;-------------------------------
		; found end of token/value
		;-------------------------------
		ODS2	<"End Token: CR.">
		Call	On_Have_Token,[@@token_start],[@@token_size]

	.ELSEIF [parse_status]==P2_HUNT_END_BR
		Call	Print_Error,offset sz_err_016,[@@token_start],[@@token_size]

	.ELSEIF [parse_status]==P2_HUNT_END_COMMENT2
		jmp	@@parse_next

	.ELSEIF [parse_status]==P2_HUNT_SEPARATOR2
		Call	Print_Error,offset sz_err_023,[@@token_start],[@@token_size]
		
	.ENDIF

	mov	[parse_status],P2_HUNT_TOKEN
	mov	[parse_mode],P2_MODE_DEFAULT

	;---------------------------
	;new line -> inc counter
	;---------------------------
	inc	[text_line_crt]
	Call	Get_Include_Context
	inc	[eax.include_line_crt]

	;ODS1	<"Line count on CR:">
	;ODV	[eax.include_line_crt]

	jmp	@@parse_next


@@Is_LF:
	;----------------------------------------------
	; this is also a separator if we hunt for it
	;----------------------------------------------
	.IF [parse_status]==P2_HUNT_SEPARATOR
		;-------------------------------
		; found end of token/value
		;-------------------------------
		ODS2	<"End Token: LF.">
		Call	On_Have_Token,[@@token_start],[@@token_size]

	.ELSEIF [parse_status]==P2_HUNT_END_COMMENT2
		jmp	@@parse_next
	.ENDIF

	mov	[parse_status],P2_HUNT_TOKEN
	mov	[parse_mode],P2_MODE_DEFAULT

	;--------------------------------------------
	;new line only if previouse char is not CR
	;aka only LF UNIX style
	;--------------------------------------------
	mov	al,[char_old]
	.IF al!=ASCII_CR
		inc	[text_line_crt]
		Call	Get_Include_Context
		inc	[eax.include_line_crt]

		;ODS1	<"Line count on LF:">
		;ODV	[eax.include_line_crt]
	.ENDIF

	jmp	@@parse_next


@@Is_Semi:
	;----------------------------------------------
	; this is also a separator if we hunt for it
	;----------------------------------------------
	.IF [parse_status]==P2_HUNT_SEPARATOR
		;-------------------------------
		; found end of value?
		;-------------------------------
		ODS2	<"End Token: Semi.">
		Call	On_Have_Token,[@@token_start],[@@token_size]
		mov	[parse_status],P2_IGNORE_NEXT
		mov	[parse_mode],P2_MODE_DEFAULT

	.ELSEIF [parse_status]==P2_HUNT_SEPARATOR2
		;-----------------------------------
		; do not comment inside a "string"
		;-----------------------------------		
	.ELSEIF [parse_status]==P2_HUNT_END_COMMENT2
		jmp	@@parse_next
	.ELSE
		;----------------
		; default action
		;-----------------
		mov	[parse_status],P2_IGNORE_NEXT
		mov	[parse_mode],P2_MODE_DEFAULT		
	.ENDIF


	jmp	@@parse_next

@@Is_Slash:
	.IF [parse_status]==P2_HUNT_TOKEN

		;-------------------------------
		;check next char for /*
		;-------------------------------
		mov	al,byte ptr [esi+1]
		.IF al=="*"
			ODS1	<"Start Block Comment:Is Slash+Star">
			mov	[parse_status],P2_HUNT_END_COMMENT2
		.ELSE
			mov	[@@token_start],esi
			mov	[@@token_size],0
			mov	[parse_status],P2_HUNT_SEPARATOR
			jmp	@@parse_next					
		.ENDIF
	.ENDIF

	jmp	@@parse_next

@@Is_Star:

	.IF [parse_status]==P2_HUNT_TOKEN
		mov	[@@token_start],esi
		mov	[@@token_size],0
		mov	[parse_status],P2_HUNT_SEPARATOR
		jmp	@@parse_next					

	.ELSEIF [parse_status]==P2_HUNT_END_COMMENT2
		mov	al,byte ptr [esi+1]
		.IF al=="/"
			ODS1	<"End Block Coment:Is Star+Slash">
			inc	esi
			inc	[@@parse_count]
			mov	[parse_status],P2_HUNT_TOKEN						
		.ENDIF
	.ENDIF	
	jmp	@@parse_next
	
@@Is_Comma:
	;----------------------------------------------
	; this is also a separator if we hunt for it
	;----------------------------------------------
	.IF [parse_status]==P2_HUNT_SEPARATOR
		;-------------------------------
		; found end of toke/value?
		;-------------------------------
		ODS2	<"End Token: Comma.">
		Call	On_Have_Token,[@@token_start],[@@token_size]
		mov	[parse_status],P2_HUNT_TOKEN
	.ENDIF

	jmp	@@parse_next

@@Is_TAB:
	.IF [parse_status]==P2_HUNT_SEPARATOR
		;-------------------------------
		; found end of token/value
		;-------------------------------
		ODS2	<"End Token: TAB.">
		Call	On_Have_Token,[@@token_start],[@@token_size]
		mov	[parse_status],P2_HUNT_TOKEN
	.ENDIF


	jmp	@@parse_next


@@Is_Equal:
	;----------------------------------------------
	; this is also a separator if we hunt for it
	;----------------------------------------------
	.IF	[parse_status]==P2_HUNT_SEPARATOR
		ODS2	<"End Token: =">
		Call	On_Have_Token,[@@token_start],[@@token_size]
		mov	[parse_status],P2_HUNT_TOKEN
	.ENDIF

	jmp	@@parse_next

@@Is_Other:
	;-------------------------
	; ignore?
	;-------------------------
	jmp	@@parse_next	

ENDP






