;------------------------------------------------------
; Sol_Asm assembler core
; Copyright (c) 2004-2005, Bogdan Valentin Ontanu.
; All rights reserved.
;------------------------------------------------------


;========================================
; Special parser for ModRM Expressions
;========================================


.data
	P3_HUNT_OP	EQU	1
	P3_HUNT_TOKEN	EQU	2

	flag_modrm_star	dd	0
.code

Parse_ModRM PROC
	USES	ecx,esi,edi
	ARG	@@token_start,@@token_size:dword

	LOCAL	@@status:dword
	LOCAL	@@tkn_start,@@tkn_size:dword


	ODS1	<"+++Parse ModRM">

	mov	esi,[@@token_start]
	mov	ecx,[@@token_size]

	xor	eax,eax
	mov	[factor_nr],eax
	mov	[flag_modrm_base],eax
	mov	[flag_modrm_index],eax
	mov	[flag_modrm_scale],eax
	mov	[flag_modrm_disp],eax

	mov	[flag_modrm_star],eax
	
	mov	[modrm_base],eax
	mov	[modrm_index],eax
	mov	[modrm_scale],eax
	mov	[modrm_disp],eax


	mov	[@@status],P3_HUNT_TOKEN
	mov	[@@tkn_size],0

@@parse_loop:
	mov	al,byte ptr [esi]

	cmp	al,"*"
	jz	@@Is_Star

	cmp	al,"+"
	jz	@@Is_Plus


	jmp	@@Is_Letter


@@next_char:
	inc	[@@tkn_size]
	inc	esi
	dec	ecx
	jnz	@@parse_loop

	;---------------------------------
	; last token has no separator :)
	;---------------------------------
	.IF [@@status]==P3_HUNT_OP
		Call	On_Modrm_Token_Plus,[@@tkn_start],[@@tkn_size]
		.IF eax==-1
			ret
		.ENDIF
	.ENDIF

	ODS1	<"Found factors#:">
	ODV	[factor_nr]

	mov	eax,[factor_nr]
	ret

@@Is_Letter:
	.IF [@@status]==P3_HUNT_TOKEN
		mov	[@@tkn_start],esi
		mov	[@@tkn_size],0
		mov	[@@status],P3_HUNT_OP
	.ENDIF
	jmp	@@next_char

@@Is_Plus:
	.IF [@@status]==P3_HUNT_OP
		Call	On_Modrm_Token_Plus,[@@tkn_start],[@@tkn_size]
		.IF eax==-1
			ret
		.ENDIF
		mov	[@@status],P3_HUNT_TOKEN
		mov	[flag_modrm_star],0
	.ELSE
		; error
		mov	eax,-1
		ret
	.ENDIF

	jmp	@@next_char

@@Is_Star:
	.IF [@@status]==P3_HUNT_OP
		Call	On_Modrm_Token_Star,[@@tkn_start],[@@tkn_size]
		.IF eax==-1
			ret
		.ENDIF
		mov	[@@status],P3_HUNT_TOKEN
		mov	[flag_modrm_star],1	
	.ELSE
		; error
		mov	eax,-1
		ret
	.ENDIF

	jmp	@@next_char
ENDP

.data
	align 4
	factor_nr		dd	0

	flag_modrm_base		dd	0
	modrm_base		dd	0

	flag_modrm_disp		dd	0
	modrm_disp		dd	0

	flag_modrm_index	dd	0
	modrm_index		dd	0

	flag_modrm_scale	dd	0
	modrm_scale		dd	0
.code


On_Modrm_Token_Plus PROC
	USES	esi,edi
	ARG	@@token_start,@@token_size:dword

	ODS1	<"+++On Modrm Token Plus">
	Call	Str_Copy_Nr,[@@token_start],offset sz_tmp1,[@@token_size]
	ODS	<offset sz_tmp1>

	Call	Check_Register,[@@token_start],[@@token_size]
	.IF eax==-1
		Call	Check_Value,[@@token_start],[@@token_size]
		.IF eax==-1
			;-------------------------
			;no register + no value
			;-------------------------
			mov	eax,-1
			ret
		.ELSE
			;-----------------------------
			;it's a number
			;-----------------------------
			.IF [flag_modrm_star]==1
				ODS1	<"Is ModRm Scale">
				mov	[modrm_scale],ebx
				mov	[flag_modrm_scale],1					
				xor	eax,eax
				ret			
			.ELSE
				.IF [flag_modrm_disp]==0
					ODS1	<"Is ModRm Displacement">
					mov	[modrm_disp],ebx		
					mov	[flag_modrm_disp],1
					inc	[factor_nr]
					xor	eax,eax
					ret
				.ELSE
					;-------------------------
					; error, 2xdisplacemnts
					;-------------------------
					mov	eax,-1
					ret
				.ENDIF
			.ENDIF
		.ENDIF
	.ELSE
		;-----------------------------
		;it's a register
		;-----------------------------
		.IF [flag_modrm_star]==1
			ODS1	<"Is ModRm Index">
			mov	[modrm_index],ebx
			mov	[flag_modrm_index],1			
			inc	[factor_nr]
			
			xor	eax,eax
			ret			
		.ENDIF
		
		.IF [flag_modrm_base]==0
			ODS1	<"Is ModRm base">
			mov	[modrm_base],ebx
			mov	[flag_modrm_base],1
			inc	[factor_nr]
			xor	eax,eax
			ret
		.ELSEIF [flag_modrm_index]==0
			ODS1	<"Is ModRm Index">
			mov	[modrm_index],ebx
			mov	[flag_modrm_index],1			
			inc	[factor_nr]
			xor	eax,eax
			ret
		.ELSE
			;----------------------------
			; error:too many registers
			;----------------------------
			mov	eax,-1
			ret
		.ENDIF
	.ENDIF

	xor	eax,eax
	ret

ENDP

On_Modrm_Token_Star PROC
	USES	esi,edi
	ARG	@@token_start,@@token_size:dword

	ODS1	<"+++On Modrm Token Star">
	Call	Str_Copy_Nr,[@@token_start],offset sz_tmp1,[@@token_size]
	ODS	<offset sz_tmp1>

	Call	Check_Value,[@@token_start],[@@token_size]
	.IF eax==-1
		Call	Check_Register,[@@token_start],[@@token_size]
		.IF eax==-1
			;-------------------------
			;no register + no value
			;-------------------------
			mov	eax,-1
			ret
		.ELSE
			;---------------------
			; is register, *
			;---------------------
			.IF [flag_modrm_index]==0
				ODS1	<"Is ModRm Index">
				mov	[modrm_index],ebx		
				mov	[flag_modrm_index],1	
				inc	[factor_nr]
				xor	eax,eax	
				ret
			.ELSE
				;-------------------------
				;error:2xindex
				;-------------------------
				mov	eax,-1
				ret
			.ENDIF
		.ENDIF
	.ELSE
		;----------------------------
		;it is a number,*
		;----------------------------
		.IF [flag_modrm_scale]==0
			ODS1	<"Is ModRm Scale">
			mov	[modrm_scale],ebx		
			mov	[flag_modrm_scale],1

			xor	eax,eax	
			ret
		.ELSE
			;--------------------------------------
			; error:expresions not implemented yet
			;--------------------------------------
			mov	eax,-1
			ret
		.ENDIF
	.ENDIF


	xor	eax,eax	
	ret

ENDP
