;------------------------------------------------------
; Sol_Asm assembler core
; Copyright (c) 2004-2005, Bogdan Valentin Ontanu.
; All rights reserved.
;------------------------------------------------------

;*****************************************
; Utils for parsing one token
;*****************************************


;===================================================
; Compare the token with a list of known keywords
;
; Warning: return into globals:
; (non reentrant without push)
; -token_type
; -token_method
; -token_params
;===================================================
.data
	align 4
	token_type	dd	0
	token_method	dd	0
	token_params	dd	0
.code

;Check_Token PROTO :dword,:dword
Check_Token PROC
	USES	ebx,ecx,edx,esi,edi
	ARG	@@token_start:dword
	ARG	@@token_size:dword


	ODS1	<"+++Check Token:">

	Call	Str_Copy_Nr,[@@token_start],offset sz_tmp1,[@@token_size]
	ODS	<offset sz_tmp1>
	ODV	[@@token_size]

	mov	edi,offset token_list

@@loop_tokens:
	mov	esi,[@@token_start]
	mov	ecx,[@@token_size]	

	movzx	edx,byte ptr [edi]	;edx==this token size

	inc	edi			;position on token string start

	;-------------------------
	;fast tests before
	;-------------------------
	cmp	edx,ecx			;compare size
	jnz	@@next_token
	

	Call	Str_Cmp_Nz_Token,esi,ecx,edi,edx

	cmp	eax,1
	jz	@@ok_match

@@next_token:
	;-------------------------------------
	; NO match this time
	; --> try next token
	;-------------------------------------
	add	edi,edx		;skip token lenght
	mov	edi,[edi]	;get pointer to next token
	test	edi,edi
	jz	@@no_match

	jmp	@@loop_tokens



@@ok_match:
	ODS2	<"Token Match OK">

	add	edi,edx		;add string length
	add	edi,4		;skip next pointer
	mov	ebx,[edi]	;get token type
	add	edi,4		;skip token type
	mov	eax,[edi]	;get procedure to execute==method

	;-------------------------
	;return global method
	;-------------------------
	mov	[token_type],ebx
	mov	[token_method],eax
	mov	[token_params],edi
	
	;-------------------------
	;return eax=0 for match
	;-------------------------
	xor	eax,eax
	ret	


@@no_match:
	ODS2	<"NO Match">
	;-------------------------------
	; return eax=-1 for NO match
	;-------------------------------
	mov	eax,-1
	ret
ENDP


;================================
; kind of specific strtok()
;================================
Peek_Line_For_Token PROC
	USES	esi,edi
	ARG	@@start_pos:dword

	LOCAL	@@tkn_start,@@tkn_size:dword
	LOCAL	@@parse_status:dword
	
	ODS1	<"+++Peek Line for Token">
	;--------------------------
	; ignore spaces and tab
	;--------------------------
@@loop_search1:
	mov	al,byte ptr [esi]
	
	cmp	al,ASCII_SPACE
	jz	@@next1
	cmp	al,ASCII_TAB
	jz	@@next1
	cmp	al,ASCII_CR
	jz	@@not_ok
	cmp	al,ASCII_LF
	jz	@@not_ok

	jmp	@@found_token_start

@@next1:
	inc	esi
	jmp	@@loop_search1
	

	;-------------------------
	; found a token start
	;-------------------------
@@found_token_start:
	mov	[@@tkn_start],esi
	mov	[@@tkn_size],0

@@loop_search2:
	mov	al,byte ptr [esi]

	cmp	al,ASCII_SPACE
	jz	@@token_end

	cmp	al,ASCII_TAB
	jz	@@token_end

	cmp	al,ASCII_CR
	jz	@@token_end

	cmp	al,ASCII_LF
	jz	@@token_end

	inc	esi
	inc	[@@tkn_size]

	jmp	@@loop_search2


@@token_end:
	xor	eax,eax	
	mov	ebx,[@@tkn_start]
	mov	ecx,[@@tkn_size]
	ret

@@not_ok:
	mov	eax,-1
	ret
ENDP



;=======================================
;Compares 2 NON null terminated strings
;case insensitive
;=======================================
Str_Cmp_Nz_Token PROC
	USES	ebx,ecx,edx,esi,edi

	ARG	@@lp_str1,@@str_len1:dword
	ARG	@@lp_str2,@@str_len2:dword


	;ODS1	<"+++Str Cmp NZ">
	;Call	Str_Copy_Nr,[@@lp_str1],offset sz_tmp1,[@@str_len1]
	;ODS	<offset sz_tmp1>
	;Call	Str_Copy_Nr,[@@lp_str2],offset sz_tmp1,[@@str_len2]
	;ODS	<offset sz_tmp1>

	mov	eax,[@@str_len1]
	mov	ebx,[@@str_len2]
	cmp	eax,ebx
	jnz	@@no_match

	mov	esi,[@@lp_str1]
	mov	edi,[@@lp_str2]
	mov	ecx,eax

@@loop_cmp:
	xor	eax,eax
	xor	ebx,ebx
	mov	al,byte ptr [esi]
	mov	bl,byte ptr [edi]

	inc	edi
	inc	esi

	cmp	al,41h		;ASCII "A"
	jl	@@not_capital

	cmp	al,5ah		;ASCII "Z"
	jg	@@not_capital

	or	al,20h		;to lower case
@@not_capital:

	cmp	eax,ebx
	jnz	@@no_match

	dec	ecx
	jnz	@@loop_cmp

	;------------------
	; return 1=Match
	;------------------
	xor	eax,eax
	inc	eax
	ret

@@no_match:
	;------------------
	; return 0=NO Match
	;------------------
	xor	eax,eax
	ret
ENDP

.data
	align 4

	list_sreg2:
		db	"es  "
		db	"cs  "
		db	"ss  "
		db	"ds  "

	list_sreg3:
		db	"es  "
		db	"cs  "
		db	"ss  "
		db	"ds  "
		db	"fs  "
		db	"gs  "

	list_regs08:
		db	"al  "
		db	"cl  "
		db	"dl  "
		db	"bl  "
		db	"ah  "
		db	"ch  "
		db	"dh  "
		db	"bh  "

	list_regs16:
		db	"ax  "
		db	"cx  "
		db	"dx  "
		db	"bx  "
		db	"sp  "
		db	"bp  "
		db	"si  "
		db	"di  "

	list_regs32:
		db	"eax "
		db	"ecx "
		db	"edx "
		db	"ebx "
		db	"esp "
		db	"ebp "
		db	"esi "
		db	"edi "

	list_regs_mmx:
		db	"mm0 "
		db	"mm1 "
		db	"mm2 "
		db	"mm3 "
		db	"mm4 "
		db	"mm5 "
		db	"mm6 "
		db	"mm7 "

	list_regs_xmmx:
		db	"xmm0"
		db	"xmm1"
		db	"xmm2"
		db	"xmm3"
		db	"xmm4"
		db	"xmm5"
		db	"xmm6"
		db	"xmm7"
.code

Check_Register PROC
	USES	ecx,esi,edi
	ARG	@@token_start,@@token_size:dword


	ODS1	<"+++Check Register">

	;Call	Str_Copy_Nr,[@@token_start],offset sz_tmp1,[@@token_size]
	;ODS	<offset sz_tmp1>
	;ODV	[@@token_size]

	mov	eax,[@@token_size]
	.IF eax<2
		;------------------------
		; any register name<2 ?
		;------------------------
		ODS1	<"---Register not Found len<2">
		mov	eax,-1
		ret

	.ELSEIF eax==2
		;-------------------------------
		; test al,cl,dl,bl,ah,ch,dh,bh
		;-------------------------------
		mov	edi,offset list_regs08
		mov	esi,[@@token_start]
		mov	ebx,[esi]
		and	ebx,00ffffh	;masks only low 2 chr al,bl,cl etc
		or	ebx,20202020h	;pad with spaces		

		;-----------------
		;reg size=8 bits
		;-----------------
		mov	eax,8
		.IF ebx==[edi+4*0]
			mov	ebx,0
			ret
		.ELSEIF ebx==[edi+4*1]
			mov	ebx,1
			ret
		.ELSEIF ebx==[edi+4*2]
			mov	ebx,2
			ret
		.ELSEIF ebx==[edi+4*3]
			mov	ebx,3
			ret
		.ELSEIF ebx==[edi+4*4]
			mov	ebx,4
			ret
		.ELSEIF ebx==[edi+4*5]
			mov	ebx,5
			ret
		.ELSEIF ebx==[edi+4*6]
			mov	ebx,6
			ret
		.ELSEIF ebx==[edi+4*7]
			mov	ebx,7
			ret
		.ENDIF

		;-------------------------------
		; test ax,cx,dx,bx,bp,sp,si,di
		;-------------------------------
		mov	edi,offset list_regs16
		mov	esi,[@@token_start]
		mov	ebx,[esi]
		and	ebx,00ffffh	;masks only low 2 chr ax,cx etc
		or	ebx,20202020h	;pad with spaces		

		;-----------------
		;reg size=16 bits
		;-----------------
		mov	eax,16
		.IF ebx==[edi+4*0]
			mov	ebx,0
			ret
		.ELSEIF ebx==[edi+4*1]
			mov	ebx,1
			ret
		.ELSEIF ebx==[edi+4*2]
			mov	ebx,2
			ret
		.ELSEIF ebx==[edi+4*3]
			mov	ebx,3
			ret
		.ELSEIF ebx==[edi+4*4]
			mov	ebx,4
			ret
		.ELSEIF ebx==[edi+4*5]
			mov	ebx,5
			ret
		.ELSEIF ebx==[edi+4*6]
			mov	ebx,6
			ret
		.ELSEIF ebx==[edi+4*7]
			mov	ebx,7
			ret
		.ENDIF

		ODS1	<"---Register not Found len=2">	
		
	.ELSEIF eax==3
		;--------------------------------------
		; test eax,ecx,edx,ebx,esp,ebp,esi,edi
		; assume reg ASCII fits in one register?
		;--------------------------------------
		mov	edi,offset list_regs32
		mov	esi,[@@token_start]
		mov	ebx,[esi]
		and	ebx,00ffffffh
		or	ebx,20202020h	;pad with spaces

		;-----------------
		;reg size=32 bits
		;-----------------
		mov	eax,32
		.IF ebx==[edi+4*0]
			mov	ebx,0
			ret
		.ELSEIF ebx==[edi+4*1]
			mov	ebx,1
			ret
		.ELSEIF ebx==[edi+4*2]
			mov	ebx,2
			ret
		.ELSEIF ebx==[edi+4*3]
			mov	ebx,3
			ret
		.ELSEIF ebx==[edi+4*4]
			mov	ebx,4
			ret
		.ELSEIF ebx==[edi+4*5]
			mov	ebx,5
			ret
		.ELSEIF ebx==[edi+4*6]
			mov	ebx,6
			ret
		.ELSEIF ebx==[edi+4*7]
			mov	ebx,7
			ret
		.ENDIF
		ODS1	<"---Register not Found len=3">	
	.ENDIF

	ODS1	<"---Register not Found2">	
	mov	eax,-1
	ret
ENDP





.data?
	align 4
	sz_expression	db	1024 dup(?)
.code
;===================================
; trim all spaces 
;===================================
Trim_Expression PROC
	USES	ecx,esi,edi
	ARG	@@token_start,@@token_size:dword

	ODS1	<"+++Trim Expression">

	mov	esi,[@@token_start]
	mov	edi,offset sz_expression
	mov	ecx,[@@token_size]
	xor	ebx,ebx			;new size


@@loop_trim:
	mov	al,byte ptr [esi]

	cmp	al,ASCII_SPACE
	jz	@@next_char
	cmp	al,ASCII_TAB
	jz	@@next_char
	;------------------------------
	; mormal chars are just copied
	;------------------------------
	mov	byte ptr [edi],al
	inc	edi
	inc	ebx

@@next_char:
	inc	esi
	dec	ecx
	jnz	@@loop_trim

	;--------------------------------
	;Return:
	;-eax=new token start
	;-ebx=new token size
	;--------------------------------
	mov	eax,offset sz_expression
	ret

ENDP





;============================================
;
; Check IF THIS token is a value/number
; Returns:
;	eax=-1 on error "not a number"
;
;	eax=0 on number ok
;	ebx=number value
;============================================
Check_Value PROC
	USES	ecx,esi,edi

	ARG	@@token_start,@@token_size:dword

	ODS1	"+++Check Value"

	Call	Str_Copy_Nr,[@@token_start],offset sz_tmp1,[@@token_size]
	ODS	<offset sz_tmp1>
	ODV	[@@token_size]

	mov	esi,[@@token_start]
	mov	ecx,[@@token_size]
	dec	ecx

	xor	eax,eax
	mov	al,byte ptr [esi+ecx]	;test last char for hints about base

	ODV	eax

	cmp	al,"h"
	jz	@@test_hexa
	cmp	al,"H"
	jz	@@test_hexa

	ODS1	<"Checking as Decimal">

	;-------------------------------
	; no hint restore original size
	;-------------------------------
	mov	ecx,[@@token_size]
@@loop_test_dec:
	mov	al,byte ptr [esi]

	cmp	al,030h			;"0" ascii
	jc	@@wrong_nr1

	cmp	al,03Ah			;"9" ascii + 1
	jnc	@@wrong_nr2
		
	inc	esi
	dec	ecx
	jnz	@@loop_test_dec

	;----------------------------
	; ok all are numbers
	; just convert them
	;----------------------------
	Call	Str_To_Val_Dec,[@@token_start],[@@token_size]
	mov	ebx,eax
	xor	eax,eax
	ret


@@test_hexa:

	ODS1	<"Checking as hexa">

@@loop_test_hex:
	xor	eax,eax
	mov	al,byte ptr [esi]

	;ODV	eax

	cmp	al,030h			;"0" ascii
	jc	@@wrong_nr1

	cmp	al,03Ah			;"9" ascii + 1
	jc	@@ok_nr_hex

	or	al,20h			;to lower case
	cmp	al,"a"
	jc	@@wrong_nr1	

	cmp	al,"f"
	ja	@@wrong_nr1

@@ok_nr_hex:		
	inc	esi
	dec	ecx
	jnz	@@loop_test_hex

	;----------------------------
	; ok all are digits
	; just convert them
	;----------------------------
	Call	Str_To_Val_Hex,[@@token_start],[@@token_size]
	mov	ebx,eax
	xor	eax,eax
	ret	

@@wrong_nr1:
	ODS1	<"Check Value-->Wrong Number1">
	mov	eax,-1
	ret

@@wrong_nr2:
	ODS1	<"Check Value-->Wrong Number2">
	mov	eax,-1
	ret

ENDP



Str_To_Val_Dec PROC
	USES	ebx,ecx,edx,esi,edi
	ARG	@@lp_str:dword
	ARG	@@str_size:dword

	ODS1	<"+++Atoi decimal">
	Call	Str_Copy_Nr,[@@lp_str],offset sz_tmp1,[@@str_size]
	ODS	<offset sz_tmp1>

	mov	esi,[@@lp_str]
	mov	edi,[@@str_size]

	mov	ebx,10		;numeric base
	xor	eax,eax		;crt value=0
	xor	edx,edx

@@loop_value:
	xor	ecx,ecx
	mov	cl,byte ptr [esi]

	mul	ebx		; N= N * 10 

	sub	cl,"0"
	add	eax,ecx	; 	N=N*10+this_digit_value

	inc	esi

	dec	edi
	jnz	@@loop_value

	;------------------------
	; eax=numerical value
	;------------------------
	ODV	eax


	ret
ENDP

Str_To_Val_Hex PROC
	USES	ebx,ecx,edx,esi,edi
	ARG	@@lp_str:dword
	ARG	@@str_size:dword

	ODS1	<"+++Atoi hexa">
	Call	Str_Copy_Nr,[@@lp_str],offset sz_tmp1,[@@str_size]
	ODS	<offset sz_tmp1>

	mov	esi,[@@lp_str]
	mov	edi,[@@str_size]
	dec	edi			;compensate last "H"

	mov	ebx,16		;numeric base
	xor	eax,eax		;crt value=0
	xor	edx,edx

@@loop_value:
	xor	ecx,ecx
	mov	cl,byte ptr [esi]

	mul	ebx		; N= N * base

	cmp	cl,"9"		;test range 0...9
	jbe	@@just_digit

	;---------------------
	; range A..F or a..f
	;---------------------
	or	cl,20h		;to lower case 
	sub	cl,"a"		;get digit value - 10
	add	cl,10		;offset it
	add	eax,ecx
	jmp	@@next_digit

@@just_digit:		
	sub	cl,"0"		;ASCII for "0" result=number0..9
	add	eax,ecx		;N=N*base+this_digit_value

@@next_digit:
	;------------------
	;next ascii char
	;------------------
	inc	esi
	dec	edi
	jnz	@@loop_value
	;------------------------
	; eax=numerical value
	;------------------------
	ODVh	eax

	ret
ENDP

