;------------------------------------------------------
; Sol_Asm assembler core
; Copyright (c) 2004-2005, Bogdan Valentin Ontanu.
; All rights reserved.
;------------------------------------------------------

;=================================
; Procedures
;=================================

PROC_STRU STRUC
	proc_name	db	128 dup(?)
	proc_name_len	dd	?
	proc_next	dd	?	;next link in hash chain

	proc_type	dd	?	;NEAR,FAR,INTER

	proc_conv	dd	?	;STDCALL,C

	proc_size	dd	?

	proc_offset	dd	?
	proc_segment	dd	?

	proc_uses_nr	dd	?
	proc_uses_regs	dd	32 dup(?)

	proc_arg_nr	dd	?
	proc_arg_size	dd	?

	proc_local_nr	dd	?
	proc_local_size	dd	?
	
ENDS

;---------------------------
;Proc types
;---------------------------
PROC_TYPE_NEAR	EQU	0
PROC_TYPE_FAR	EQU	1
PROC_TYPE_INTER	EQU	2


PROC_MAX	EQU	128*1024

.data
	proc_count	dd	-1
	proc_crt	dd	-1
.data?
	PROC_MAX	EQU		128*1024
	proc_base0	PROC_STRU	PROC_MAX dup (<?>)
	proc_hash0	dd		65538 dup(?)
.code
Clear_Hash_Procs PROC
	USES	ecx,edi

	ODS1	<"+++Clear Hash Procs">

	mov	edi,offset proc_hash0
	mov	ecx,65537
	mov	eax,-1

@@loop_clear:
	mov	[edi],eax
	add	edi,4
	dec	ecx
	jnz	@@loop_clear

	ret
ENDP

;Create_Proc PROTO :dword, :dword, :dword
Create_Proc PROC
	USES	esi,edi
	ARG 	@@token_start,@@token_size:dword
	ARG	@@proc_offset:dword

	ODS1	<"+++Create PROC">

	Call	Str_Copy_Nr,[@@token_start],offset sz_tmp1,[@@token_size]
	ODS	<offset sz_tmp1>

	mov	eax,[proc_count]
	inc	eax
	mov	[proc_count],eax

	cmp	eax,PROC_MAX
	jc	@@proc_ok

	dec	eax
	mov	[proc_count],eax
	;-------------------------------------
	; Error: can not create more proc's
	;-------------------------------------
	Call	Print_Error,offset sz_err_024,[@@token_start],[@@token_size]
	mov	eax,-1
	ret

@@proc_ok:
	;--------------------------------
	; we are in this procedure body
	;--------------------------------
	mov	[proc_crt],eax
	;------------------------------
	;calculate the address
	;------------------------------
	mov	esi,offset proc_base0
	mov	ebx,SIZE PROC_STRU
	mul	ebx
	add	esi,eax

	push	esi
		add	esi,proc_name
		Call	Str_Copy_Nr,[@@token_start],esi,[@@token_size]
	pop	esi
	
	mov	eax,[@@token_size]
	mov	[esi.proc_name_len],eax

	mov	eax,[@@proc_offset]
	mov	[esi.proc_offset],eax

	; init stuff to zero
	xor	eax,eax
	mov	[esi.proc_uses_nr],eax

	mov	[esi.proc_arg_nr],eax
	mov	[esi.proc_arg_size],eax

	mov	[esi.proc_local_nr],eax
	mov	[esi.proc_local_size],eax

	;----------------------------
	;make the hash value
	;----------------------------
	push	esi
	mov	edi,[@@token_start]
	mov	esi,[@@token_size]
	Call	Make_Hash
	pop	esi

	;-----------------------
	; ebx=hash value
	;-----------------------
	shl	ebx,2		;*4
	mov	edi,offset proc_hash0
	add	edi,ebx
	mov	eax,[edi]

	ODS1	<"id at hash:">
	ODVh	eax	
	.IF eax==-1
		;---------------------------
		; lucky = empty hash location
		;----------------------------
		ODS1	<"Proc:Add on empty hash">
		mov	ebx,[proc_count]
		mov	[edi],ebx
		mov	[esi.proc_next],eax
	.else
		;---------------------------------
		; colission, eax=old guy
		;---------------------------------
		ODS1	<"Proc:Add collission hash">
		mov	ebx,[proc_count]
		mov	[edi],ebx
		mov	[esi.proc_next],eax
	.endif
	
	mov	eax,esi
	ret
ENDP

Generate_Prologue PROC
	USES	ebx,ecx,esi,edi

	ODS1	<"+++Generate Prologue:">
	ODS2	<" Proc id is:">
	ODV	[proc_crt]

	mov	eax,[proc_crt]
	.IF eax==-1
		ret
	.ENDIF


	
	;-------------------------
	;calc base addr
	;-------------------------
	mov	eax,[proc_crt]
	mov	esi,offset proc_base0
	;mov	ecx,SIZE PROC_STRU
	;mul	ecx
	;add	esi,eax
	
	;-------------------------------------
	; ENTER
	; encode using level==0 always?
	;-------------------------------------
	mov	eax,[esi.proc_local_size]
	Call	Encode_ENTER,eax,0
	
	mov	eax,[esi.proc_uses_nr]
	.IF eax==0
		ODS1	<"PROC has NO Uses">
		jmp	@@after_push
	.ELSE
		ODS1	<"PROC Uses:">
		ODV	eax
	.ENDIF

	;-------------------------
	;push needed registers
	;-------------------------
	mov	ebx,eax
	xor	ecx,ecx

@@loop_push:
	
	mov	eax,[esi.proc_uses_regs+ecx]
	pushad
	Call	Encode_PUSH_Reg1,eax
	popad

	add	ecx,4
	dec	ebx
	jnz	@@loop_push

@@after_push:


	ret
ENDP

Generate_Epilogue PROC
	USES	ebx,ecx,esi,edi

	ODS1	<"+++Generate Epilogue">
	ODS2	<" Proc id is:">
	ODV	[proc_crt]

	mov	eax,[proc_crt]
	.IF eax==-1
		ret
	.ENDIF
	
	;-------------------------
	;calc base addr
	;-------------------------
	mov	esi,offset proc_base0
	mov	ebx,SIZE PROC_STRU
	mul	ebx
	add	esi,eax

	mov	eax,[esi.proc_uses_nr]
	.IF eax==0
		ODS1	<"PROC has NO Uses">
		jmp	@@after_pop
	.ELSE
		ODS1	<"PROC Uses:">
		ODV	eax
	.ENDIF

	;-------------------------
	;loop to pop registers
	;-------------------------
	mov	ebx,eax
	mov	ecx,eax
	dec	ecx
	shl	ecx,2

@@loop_pop:
	
	mov	eax,[esi.proc_uses_regs+ecx]
	pushad
	Call	Encode_POP_Reg1,eax
	popad

	sub	ecx,4
	dec	ebx
	jnz	@@loop_pop
	

@@after_pop:

	Call	Encode_LEAVE

	mov	eax,[esi.proc_arg_size]
	.IF eax==0
		Call	Encode_RET
	.ELSE
		Call	Encode_RETN,eax
	.ENDIF

	ret
ENDP

;=========================================
; Verify IF this PROC already exists
; using a hash to speed up
;=========================================
align 4
Check_Proc PROC
	USES	esi,edi
	ARG	@@token_start,@@token_size:dword


	ODS1	<"+++Proc:Check via Hash">

	Call	Str_Copy_Nr,[@@token_start],offset sz_tmp1,[@@token_size]
	ODS	<offset sz_tmp1>
	ODV	[@@token_size]

	mov	ecx,[proc_count]
	cmp	ecx,-1
	jz	@@no_procs

	;--------------------------------
	;make hash value
	;--------------------------------
	push	esi
	mov	edi,[@@token_start]
	mov	esi,[@@token_size]
	Call	Make_Hash
	pop	esi
	;-----------------------
	; ebx=hash
	;-----------------------
	ODVh	ebx

	shl	ebx,2
	mov	edi,offset proc_hash0
	add	edi,ebx
	mov	eax,[edi]

	ODS1	<"PROC:ID at hash:">
	ODV	eax

	.IF eax==-1
		;-----------------------
		;empty hash location
		;-----------------------
		ODS1	<"PROC Not found, empty hash">
		mov	eax,-1
		ret
	.ELSE
		;------------------------------
		;hash hit, we have to check
		;------------------------------
		ODS1	<"PROC:Hash Hit">

	@@follow_link:
		ODS1	<"PROC:ID tested:">
		ODV	eax	

		mov	ebx, SIZE PROC_STRU
		mul	ebx
		mov	edi,offset proc_base0
		add	edi,eax

		mov	eax,[edi.proc_name_len]
		cmp	eax,[@@token_size]
		jnz	@@next_link

		ODS1	<"PROC:Name Length is Equal">

		push	edi

		mov	esi,[@@token_start]
		add	edi,proc_name		
		mov	ecx,[@@token_size]

		@@loop_cmp:
			mov	al,byte ptr [esi]
			mov	bl,byte ptr [edi]
			cmp	al,bl
			jnz	@@no_match

			inc	edi
			inc	esi

			dec	ecx
			jnz	@@loop_cmp

			ODS1	<"PROC: Found">

			pop	edi
			mov	eax,[edi.proc_size]	;size 4:code and dd, 2:word,1:byte
			mov	ebx,[edi.proc_offset]	;OFFSET
			ret			

		@@no_match:
			ODS1	<"PROC:Char not match">
			
			pop	edi

		@@next_link:
			mov	eax,[edi.proc_next]
			
			cmp	eax,-1
			jnz	@@follow_link	
			
			mov	eax,-1
			ret
	.ENDIF

@@no_procs:
	mov	eax,-1
	ret
ENDP


;====================================
; Verify if this Procedure exists
;====================================
Check_Proc1 PROC
	USES	esi,edi
	ARG	@@token_start,@@token_size:dword
	LOCAL	@@proc_crt:dword

	ODS1	<"+++Check PROC">

	mov	edi,offset proc_base0

	mov	ecx,[proc_count]
	cmp	ecx,-1
	jz	@@no_proc

	inc	ecx
	mov	[@@proc_crt],0

@@loop_tokens:
	mov	eax,[edi.proc_name_len]
	cmp	eax,[@@token_size]
	jnz	@@next_proc

	ODS1	<"Proc:Length is Equal">

	mov	esi,[@@token_start]
	mov	edx,[@@token_size]

	push	edi
	add	edi,proc_name
@@loop_cmp:
	mov	al,byte ptr [esi]		
	mov	bl,byte ptr [edi]
	cmp	al,bl
	jnz	@@no_match
	
	inc	edi
	inc	esi
	dec	edx
	jnz	@@loop_cmp

	pop	edi

	ODS1	<"Proc Found">
	mov	eax,0
	mov	ebx,[edi.proc_offset]
	mov	ecx,[@@proc_crt]

	ret

@@no_match:
	;ODS1	<"No match char">	
	pop	edi

@@next_proc:	
	;---------------
	; next procedure
	;---------------
	add	edi,SIZE PROC_STRU
	inc	[@@proc_crt]
	dec	ecx
	jnz	@@loop_tokens

@@no_proc:
	ODS1	<"Proc Not found">
	mov	eax,-1
	ret
ENDP


;=================================
; Util function
;=================================
.data
	sz_info_proc	db	13,10,"Printing PROCs:",13,10
			db	"-----------------------------",0

	sz_info_proc2	db	13,10
			db	"-----------------------------",13,10,0

	sz_fmt_proc	db	13,10,"[%i]%s",09,"Addr:%i",0
.code
Print_Procs PROC
	USES	esi,edi
	LOCAL	@@tmp_cnt:dword

	ODS1	<"+++Print Proc">

	.IF [proc_count]==-1
		ret
	.ENDIF

	Call	Print_Text,offset sz_info_proc

	mov	[@@tmp_cnt],0
	mov	esi,offset proc_base0
	mov	ecx,[proc_count]
	cmp	ecx,-1
	jz	@@end_procs

	inc	ecx

@@loop_procs:
	push	ecx
	push	esi
	
	mov	eax,[esi.proc_name_len]
	Call	Str_Copy_Nr,esi,offset sz_tmp1,eax

	mov	eax,[esi.proc_offset]
	Call	wsprintfA,offset sz_tmp2,offset sz_fmt_proc,[@@tmp_cnt],offset sz_tmp1,eax
	Call	Print_Text,offset sz_tmp2

	pop	esi
	pop	ecx
	;---------------
	; next proc
	;---------------
	inc	[@@tmp_cnt]
	add	esi,SIZE PROC_STRU
	dec	ecx
	jnz	@@loop_procs

@@end_procs:
	Call	Print_Text,offset sz_info_proc2

	ret
ENDP
