;********************************************
;	SOL_ASM Assembler
;********************************************

;------------------------------------------------------
; Sol_Asm assembler core
; Copyright (c) 2004-2005, Bogdan Valentin Ontanu.
; All rights reserved.
;------------------------------------------------------


;----------------------------------------
;Notes:	
;	-TABS=hard tab 8 columns
;	-to be assembled with TASM v5.0+
;	-windows version only
;----------------------------------------


.386				; minimum processor needed	
.model flat, stdcall		; FLAT memory model & STDCALL calling

LOCALS
OPTION SCOPED

;----------------------------------------------
; command for the linker
;----------------------------------------------
includelib libs\import32.lib


.data
	main_module_handle	dd	0
	std_out_handle		dd	0
	std_in_handle		dd	0


	read_bytes		dd	0
	nr_bytes_written	dd	0



	sz_msg_001		db	13,10,"Sol_Asm Assembler v0.03"
				db	13,10,"=======================",0
				
	sz_msg_help		db	13,10,"-----------------------------------------------------------"
				db	13,10,"Usage: sol_asm [-option] <input_filename> <output_filename>"
				db	13,10,32,">>>Options are:"
				db	13,10,09,"-h, -help 	==> print this help"
				db	13,10,09,"-show_labels	==> show defined labels"
				db	13,10,09,"-show_procs 	==> show defined procedures"
				db	13,10,09,"-show_strucs 	==> show defined structures"				
				db	13,10,09,"-show_files 	==> show parsed files"
				db	13,10,09,"-no_errors  	==> do not report errors"
				db	13,10,09,"-out_text   	==> redir messages to out.txt"		
				db	13,10,"-----------------------------------------------------------",0



.data?
	lp_read_buffer		db	1024 dup(?)
.code

;include equ_file.asm


;------------------------------
; includes
;------------------------------
include	win32\win32api.asm
include util\utils.asm
include util\strings.asm

include cmd_line\cmd_line_main.asm

include parser\parse_main.asm
include	parser\parse_token.asm
include	parser\list_tokens.asm
include parser\parser_messages.asm
include parser\on_have_token.asm
include parser\on_have_token_string.asm
include parser\on_have_token_br.asm
include parser\parse_modrm.asm

;-------------------------
;symbols
;-------------------------
include labels\labels.asm
include procedures\procedures.asm
include structures\structures.asm
include structures\stru_members.asm

include params\params.asm

;------------------------------
;encodings
;------------------------------
include encodings\encode_simple.asm
include encodings\encode_call.asm
include encodings\encode_jmp.asm
include encodings\encode_pop.asm
include encodings\encode_push.asm
include encodings\encode_ret.asm
include encodings\encode_mov.asm

;**************************************
;	MAIN PROGRAM ENTRY POINT
;	linker will jump us here
;**************************************
.data
	sz_fname_ods		db	"ods.txt",0
	file_handle_ods		dd	0

	sz_time_fmt		db	"---------------------",13,10
					db	"time_start:%d",13,10
					db	"time_end  :%d",13,10
					db	"duration  :%dms",13,10
					db	"---------------------",13,10,0

	time_start		dd	0
	time_end		dd	0	

.code
Start:
	;---------------------------------------------
	;   get the module handle for this instance
	;---------------------------------------------
	Call	GetModuleHandleA,0
	mov	[main_module_handle], eax	; and store it

	Call	GetTickCount
	mov	[time_start],eax

	Call	CreateFileA,offset sz_fname_ods,GENERIC_WRITE,0,0,CREATE_ALWAYS,80h,0
	mov	[file_handle_ods],eax

	Call	GetCommandLineA
	mov	[main_lp_cmd_line],eax
	

	;------------------------------------
	; Init STDIN & STDOUT streams
	;------------------------------------	
	Call	GetStdHandle,STD_INPUT_HANDLE
	mov	[std_in_handle],eax

	Call	GetStdHandle,STD_OUTPUT_HANDLE
	mov	[std_out_handle],eax


	;-------------------------------
	; Greetings Messages
	;-------------------------------
	Call	Print_Text,offset sz_msg_001


	;--------------------------------
	; check command line
	;--------------------------------
	Call	Parse_Cmd_Line,[main_lp_cmd_line]
	
	.IF [flag_option_help]==1
		;------------------------------------
		; help message
		;------------------------------------
		Call	Print_Text,offset sz_msg_help
		Call	ReadConsoleA,[std_in_handle],offset lp_read_buffer,128,offset read_bytes,0
		jmp	End_App
	.endif

	Call	Print_Text,offset sz_crlf
	Call	Print_Text,offset sz_filename_in

	;Call	Print_Text,offset sz_crlf
	;Call	Print_Text,offset sz_filename_out

	;-------------------------------
	; parse the file
	;-------------------------------
	.IF [flag_filename_in]==1
		Call	Clear_Hash_Labels
		Call	Clear_Hash_Procs
		Call	Clear_Hash_Structure
		Call	Clear_Hash_Stru_Members
		Call	Parser_Main
	.ELSE
		Call	Print_Text,offset sz_no_filename_in
		Call	Print_Text,offset sz_msg_help
		Call	ReadConsoleA,[std_in_handle],offset lp_read_buffer,128,offset read_bytes,0
		jmp	End_App
	.ENDIF


End_App:
	;---------------------------------
	; exit app, the end
	;---------------------------------
	Call	GetTickCount
	mov	[time_end],eax
	sub	eax,[time_start]
	call	wsprintfA,offset sz_tmp1,offset sz_time_fmt,[time_start],[time_end],eax
	Call	Print_Text,offset sz_tmp1

	Call	CloseHandle,[file_handle_ods]
	;Call	ReadConsoleA,[std_in_handle],offset lp_read_buffer,128,offset read_bytes,0
	Call	ExitProcess,0
ret


;******************************
;  END OF PROGRAM TEXT
;******************************
END Start
